# -*- coding: utf-8 -*-
import os

from qtpy import QtCore

from ..widgets.glassure_widget import GlassureWidget
from ..widgets.custom.file_dialogs import open_file_dialog
from ..model.glassure_model import GlassureModel


class TransferFunctionController(object):
    def __init__(self, widget, glassure_model):
        """
        :param widget:
        :type widget: GlassureWidget
        :param glassure_model:
        :type glassure_model: GlassureModel
        """

        self.widget = widget
        self.transfer_widget = widget.transfer_widget
        self.model = glassure_model
        self.settings = QtCore.QSettings('Glassure', 'Glassure')

        self.connect_signals()

    def connect_signals(self):
        self.transfer_widget.activate_cb.stateChanged.connect(self.active_cb_state_changed)

        self.transfer_widget.load_sample_btn.clicked.connect(self.load_sample_pattern)
        self.transfer_widget.load_sample_bkg_btn.clicked.connect(self.load_sample_bkg_pattern)
        self.transfer_widget.load_std_btn.clicked.connect(self.load_std_pattern)
        self.transfer_widget.load_std_bkg_btn.clicked.connect(self.load_std_bkg_pattern)

        self.transfer_widget.sample_bkg_scaling_sb.valueChanged.connect(self.sample_bkg_scaling_changed)
        self.transfer_widget.std_bkg_scaling_sb.valueChanged.connect(self.std_bkg_scaling_changed)
        self.transfer_widget.smooth_sb.valueChanged.connect(self.smooth_factor_changed)

    def load_sample_pattern(self):
        filename = open_file_dialog(self.widget, caption="Load Sample Pattern (in Container)",
                                    directory=self.settings.value('working_directory'))

        if filename != '':
            self.model.load_transfer_sample_pattern(filename)
            self.working_directory = os.path.dirname(filename)
            self.transfer_widget.sample_filename_lbl.setText(os.path.basename(filename))

    def load_sample_bkg_pattern(self):
        filename = open_file_dialog(self.widget, caption="Load Sample background Pattern (in Container)",
                                    directory=self.settings.value('working_directory'))

        if filename != '':
            self.model.load_transfer_sample_bkg_pattern(filename)
            self.working_directory = os.path.dirname(filename)
            self.transfer_widget.sample_bkg_filename_lbl.setText(os.path.basename(filename))

    def load_std_pattern(self):
        filename = open_file_dialog(self.widget, caption="Load Standard Pattern (in Container)",
                                    directory=self.settings.value('working_directory'))

        if filename != '':
            self.model.load_transfer_std_pattern(filename)
            self.working_directory = os.path.dirname(filename)
            self.transfer_widget.std_filename_lbl.setText(os.path.basename(filename))

    def load_std_bkg_pattern(self):
        filename = open_file_dialog(self.widget, caption="Load Standard Background Pattern (in Container)",
                                    directory=self.settings.value('working_directory'))

        if filename != '':
            self.model.load_transfer_std_bkg_pattern(filename)
            self.working_directory = os.path.dirname(filename)
            self.transfer_widget.std_bkg_filename_lbl.setText(os.path.basename(filename))

    def active_cb_state_changed(self):
        self.model.use_transfer_function = self.transfer_widget.activate_cb.isChecked()

    def sample_bkg_scaling_changed(self, new_value):
        self.model.transfer_sample_bkg_scaling = float(new_value)

    def std_bkg_scaling_changed(self, new_value):
        self.model.transfer_std_bkg_scaling = float(new_value)

    def smooth_factor_changed(self, new_value):
        self.model.transfer_function_smoothing = new_value



# -*- coding: utf-8 -*-

from ..widgets.glassure_widget import GlassureWidget
from ..model.glassure_model import GlassureModel


class SollerController(object):
    def __init__(self, widget, glassure_model):
        """
        :param widget:
        :type widget: GlassureWidget
        :param glassure_model:
        :type glassure_model: GlassureModel
        """

        self.widget = widget
        self.soller_widget = widget.soller_widget
        self.model = glassure_model

        self.connect_signals()

    def connect_signals(self):
        self.soller_widget.soller_parameters_changed.connect(self.parameters_changed)
        self.soller_widget.activate_cb.stateChanged.connect(self.active_cb_state_changed)

    def parameters_changed(self):
        self.model.soller_parameters = self.soller_widget.get_parameters()

    def active_cb_state_changed(self):
        self.model.use_soller_correction = self.soller_widget.activate_cb.isChecked()



# -*- coding: utf-8 -*-

import os
from qtpy import QtGui, QtWidgets, QtCore

from ..widgets.glassure_widget import GlassureWidget
from ..model.glassure_model import GlassureModel
from ..widgets.custom.file_dialogs import save_file_dialog, open_file_dialog


class ConfigurationController(object):
    def __init__(self, main_widget: GlassureWidget,
                 glassure_model: GlassureModel):
        """
        :param main_widget:
        :param glassure_model:
        """

        self.widget = main_widget
        self.model = glassure_model
        self.settings = QtCore.QSettings('Glassure', 'Glassure')

        self.connect_signals()

        self.update_configurations_tw()
        self.update_widget_controls()

    def connect_signals(self):
        self.widget.freeze_configuration_btn.clicked.connect(self.model.add_configuration)
        self.widget.remove_configuration_btn.clicked.connect(self.model.remove_configuration)
        self.widget.configuration_tw.currentCellChanged.connect(self.model.select_configuration)

        self.model.configurations_changed.connect(self.update_configurations_tw)
        self.model.configurations_changed.connect(self.update_pattern_items)
        self.model.configurations_changed.connect(self.update_widget_controls)

        self.model.configuration_selected.connect(self.update_widget_controls)
        self.model.configuration_selected.connect(self.update_pattern_items)

        self.widget.configuration_widget.configuration_show_cb_state_changed.connect(
            self.update_configuration_visibility)
        self.widget.configuration_widget.configuration_color_btn_clicked.connect(self.configuration_color_btn_clicked)
        self.widget.configuration_widget.configuration_name_changed.connect(self.update_configuration_name)
        self.widget.configuration_widget.save_btn.clicked.connect(self.save_model)
        self.widget.configuration_widget.load_btn.clicked.connect(self.load_model)

    def freeze_configuration(self):
        """
        Callback for the freeze configuration button. Adds a new configuration
        to the model with the current settings. (i.e. duplicates the current
        configuration)
        """
        self.model.add_configuration()

    def update_configurations_tw(self):
        """Updates the configuration table widget"""
        self.widget.configuration_tw.blockSignals(True)
        self.widget.configuration_widget.clear_configuration_tw()

        for configuration in self.model.configurations:
            color = configuration.color
            self.widget.configuration_widget.add_configuration(
                configuration.name,
                '#%02x%02x%02x' % (int(color[0]), int(color[1]), int(color[2])),
                show=configuration.show
            )
        self.widget.configuration_tw.blockSignals(False)
        self.widget.configuration_widget.select_configuration(self.model.configuration_ind)

    def update_widget_controls(self):
        self.widget.right_control_widget.optimization_widget.blockSignals(True)

        # filenames
        self.widget.data_filename_lbl.setText(self.model.original_pattern.name)
        if self.model.current_configuration.background_pattern is not None:
            self.widget.bkg_filename_lbl.setText(self.model.current_configuration.background_pattern.name)
            self.widget.bkg_scaling_sb.setEnabled(True)
            self.widget.bkg_scaling_sb.setValue(self.model.current_configuration.background_pattern.scaling)
        else:
            self.widget.bkg_filename_lbl.setText('None')
            self.widget.bkg_scaling_sb.setEnabled(False)
        self.widget.smooth_sb.setValue(self.model.original_pattern.smoothing)

        self.widget.update_sample_config(self.model.sample)
        self.widget.update_transform_config(self.model.transform_config)
        self.widget.left_control_widget.extrapolation_widget.update_configuration(self.model.extrapolation_config)

        # optimizations widget
        self.widget.optimize_activate_cb.setChecked(self.model.optimize)
        self.widget.set_optimization_parameter(self.model.r_cutoff,
                                               self.model.optimization_iterations,
                                               self.model.optimization_attenuation)
        self.widget.right_control_widget.optimization_widget.blockSignals(False)

        # soller widget
        self.widget.soller_active_cb.setChecked(self.model.use_soller_correction)
        self.widget.set_soller_parameter(self.model.soller_parameters)

    def update_pattern_items(self):
        self.update_data_plot()
        self.update_plot_item_count(len(self.model.configurations))
        self.update_pattern_items_data(self.model.configuration_ind)
        self.update_pattern_items_color(self.model.configuration_ind)

    def update_data_plot(self):
        self.widget.pattern_widget.plot_pattern(self.model.original_pattern)
        self.widget.pattern_widget.plot_bkg(self.model.background_pattern)

    def update_plot_item_count(self, count):
        """ Updates the number of plot items in the sq-plot and the gr-plot """

        for i in range(len(self.widget.pattern_widget.sq_items), count):
            self.widget.pattern_widget.add_sq_item(show=self.model.configurations[i].show)
            self.widget.pattern_widget.add_gr_item(show=self.model.configurations[i].show)

        while len(self.widget.pattern_widget.sq_items) > count:
            self.widget.pattern_widget.remove_sq_item()
            self.widget.pattern_widget.remove_gr_item()

    def update_pattern_items_data(self, cur_ind):
        for ind in range(len(self.model.configurations)):
            if self.model.configurations[ind].sq_pattern is None:
                continue
            self.widget.pattern_widget.set_sq_pattern(self.model.configurations[ind].sq_pattern, ind)
            self.widget.pattern_widget.set_gr_pattern(self.model.configurations[ind].gr_pattern, ind)

    def update_pattern_items_color(self, cur_ind):
        for ind in range(len(self.model.configurations)):
            if ind == self.model.configuration_ind:
                self.widget.pattern_widget.activate_ind(ind)
            else:
                self.widget.pattern_widget.set_color(self.model.configurations[ind].color, ind)

    def update_configuration_visibility(self, ind, visible):
        self.model.configurations[ind].show = visible
        if visible:
            self.widget.pattern_widget.show_sq(ind)
            self.widget.pattern_widget.show_gr(ind)
        else:
            self.widget.pattern_widget.hide_sq(ind)
            self.widget.pattern_widget.hide_gr(ind)

    def configuration_color_btn_clicked(self, ind, button):
        """
        Callback for the color buttons in the configuration table. Opens up a
        color dialog. The color of the configuration and its respective button
        will be changed according to the selection

        :param ind: configuration ind
        :param button: button to color
        """
        previous_color = button.palette().color(QtGui.QPalette.Button)
        new_color = QtWidgets.QColorDialog.getColor(previous_color, self.widget)

        if not new_color.isValid():
            return

        self.model.configurations[ind].color = [new_color.red(), new_color.green(), new_color.blue()]
        self.update_pattern_items_color(self.model.configuration_ind)
        button.setStyleSheet('background-color:' + new_color.name())

    def update_configuration_name(self, ind, name):
        self.model.configurations[ind].name = name

    def save_model(self):
        filename = save_file_dialog(
            self.widget,
            'Save model',
            self.settings.value('working_directory'),
            filter='*.json')
        if filename == '':
            return
        self.settings.setValue('working_directory', os.path.dirname(filename))
        self.model.to_json(filename)

    def load_model(self):
        filename = open_file_dialog(
            self.widget,
            'Load model',
            self.settings.value('working_directory'),
            filter='*.json')
        if filename == '':
            return
        self.settings.setValue('working_directory', os.path.dirname(filename))
        self.model.read_json(filename)



# -*- coding: utf-8 -*-
from __future__ import annotations
import os

import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter1d


class Pattern(object):
    """
    A Pattern is a set of x and y values.
    It can be loaded from a file or created from scratch and can be modified by
    different methods.
    It builds the basis for all calculations in glassure.

    :param x: x values of the pattern
    :param y: y values of the pattern
    :param name: name of the pattern
    """

    def __init__(self, x: np.ndarray = None, y: np.ndarray = None, name: str = ''):
        """
        Creates a new Pattern object, x and y should have the same shape.
        """
        if x is None:
            self._x = np.linspace(0.1, 15, 100)
        else:
            self._x = x
        if y is None:
            self._y = np.log(self._x ** 2) - (self._x * 0.2) ** 2
        else:
            self._y = y
        self.name = name
        self.offset = 0.0
        self._scaling = 1.0
        self.smoothing = 0.0
        self.bkg_pattern = None

    def load(self, filename: str, skiprows: int = 0):
        """
        Loads a pattern from a file. The file can be either a .xy or a .chi file. The .chi file will be loaded with
        skiprows=4 by default.

        :param filename: path to the file
        :param skiprows: number of rows to skip when loading the data (header)
        """
        try:
            if filename.endswith('.chi'):
                skiprows = 4
            data = np.loadtxt(filename, skiprows=skiprows)
            self._x = data.T[0]
            self._y = data.T[1]
            self.name = os.path.basename(filename).split('.')[:-1][0]

        except ValueError:
            print('Wrong data format for pattern file! - ' + filename)
            return -1

    @staticmethod
    def from_file(filename: str, skip_rows: int = 0) -> Pattern | '-1':
        """
        Loads a pattern from a file. The file can be either a .xy or a .chi file. The .chi file will be loaded with
        skiprows=4 by default.

        :param filename: path to the file
        :param skip_rows: number of rows to skip when loading the data (header)
        """
        try:
            if filename.endswith('.chi'):
                skip_rows = 4
            data = np.loadtxt(filename, skiprows=skip_rows)
            x = data.T[0]
            y = data.T[1]
            name = os.path.basename(filename).split('.')[:-1][0]
            return Pattern(x, y, name)

        except ValueError:
            print('Wrong data format for pattern file! - ' + filename)
            return -1

    def save(self, filename: str, header: str = ''):
        """
        Saves the Pattern to a two-column xy file.

        :param filename: path to the file
        :param header: header to be written to the file
        """
        data = np.dstack((self._x, self._y))
        np.savetxt(filename, data[0], header=header)

    def set_background(self, pattern: Pattern):
        """
        Sets a background pattern to the current pattern. The background will be subtracted from the current pattern
        when calling the data property.

        :param pattern: Pattern to be used as background
        """
        self.bkg_pattern = pattern

    def reset_background(self):
        """
        Resets the background pattern to None.
        """
        self.bkg_pattern = None

    def set_smoothing(self, amount: float):
        """
        Sets the smoothing amount for the pattern. The smoothing will be applied when calling the data property.

        :param amount: amount of smoothing to be applied
        """
        self.smoothing = amount

    def rebin(self, bin_size: float) -> Pattern:
        """
        Returns a new pattern, which is a rebinned version of the current one.

        :param bin_size: Size of the bins
        :return: rebinned Pattern
        """
        x, y = self.data
        x_min = np.round(np.min(x) / bin_size) * bin_size
        x_max = np.round(np.max(x) / bin_size) * bin_size
        new_x = np.arange(x_min, x_max + 0.1 * bin_size, bin_size)

        bins = np.hstack((x_min - bin_size * 0.5, new_x + bin_size * 0.5))
        new_y = (np.histogram(x, bins, weights=y)
                 [0] / np.histogram(x, bins)[0])

        return Pattern(new_x, new_y)

    @property
    def data(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns the data of the pattern. If a background pattern is set, the background will be subtracted from the
        pattern. If smoothing is set, the pattern will be smoothed.

        :return: Tuple of x and y values
        """
        if self.bkg_pattern is not None:
            # create background function
            x_bkg, y_bkg = self.bkg_pattern.data

            if not np.array_equal(x_bkg, self._x):
                # the background will be interpolated
                f_bkg = interp1d(x_bkg, y_bkg, kind='linear')

                # find overlapping x and y values:
                ind = np.where((self._x <= np.max(x_bkg)) &
                               (self._x >= np.min(x_bkg)))
                x = self._x[ind]
                y = self._y[ind]

                if len(x) == 0:
                    # if there is no overlapping between background and pattern, raise an error
                    raise BkgNotInRangeError(self.name)

                y = y * self._scaling + self.offset - f_bkg(x)
            else:
                # if pattern and bkg have the same x basis we just delete y-y_bkg
                x, y = self._x, self._y * self._scaling + self.offset - y_bkg
        else:
            x, y = self.original_data

        if self.smoothing > 0:
            y = gaussian_filter1d(y, self.smoothing)
        return x, y

    @data.setter
    def data(self, data: tuple[np.ndarray, np.ndarray]):
        """
        Sets the data of the pattern. Also resets the scaling and offset to 1 and 0 respectively.

        :param data: tuple of x and y values
        """
        (x, y) = data
        self._x = x
        self._y = y
        self.scaling = 1.0
        self.offset = 0

    @property
    def original_data(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns the original data of the pattern without any background subtraction or smoothing.

        :return: tuple of x and y values
        """
        return self._x, self._y * self._scaling + self.offset

    @property
    def x(self) -> np.ndarray:
        """ Returns the x values of the pattern """
        return self._x

    @x.setter
    def x(self, new_value: np.ndarray):
        """ Sets the x values of the pattern """
        self._x = new_value

    @property
    def y(self) -> np.ndarray:
        """ Returns the y values of the pattern """
        return self._y

    @y.setter
    def y(self, new_y: np.ndarray):
        """ Sets the y values of the pattern """
        self._y = new_y

    @property
    def scaling(self) -> float:
        """ Returns the scaling of the pattern """
        return self._scaling

    @scaling.setter
    def scaling(self, value):
        """ 
        Sets the scaling of the pattern, if below 0, it will be set to 0 
        instead.
        """
        if value < 0:
            self._scaling = 0.0
        else:
            self._scaling = value

    def limit(self, x_min: float, x_max: float) -> Pattern:
        """
        Limits the pattern to a specific x-range. Does not modify inplace but returns a new limited Pattern

        :param x_min: lower limit of the x-range
        :param x_max: upper limit of the x-range
        :return: limited Pattern
        """
        x, y = self.data
        return Pattern(x[np.where((x_min < x) & (x < x_max))],
                       y[np.where((x_min < x) & (x < x_max))])

    def extend_to(self, x_value: float, y_value: float) -> Pattern:
        """
        Extends the current pattern to a specific x_value by filling it with the y_value. Does not modify inplace but
        returns a new filled Pattern

        :param x_value: Point to which extending the pattern should be smaller than the lowest x-value in the pattern or
        vice versa
        :param y_value: number to fill the pattern with
        :return: extended Pattern
        """
        x_step = np.mean(np.diff(self.x))
        x_min = np.min(self.x)
        x_max = np.max(self.x)
        if x_value < x_min:
            x_fill = np.arange(x_min - x_step, x_value -
                               x_step * 0.5, -x_step)[::-1]
            y_fill = np.zeros(x_fill.shape)
            y_fill.fill(y_value)

            new_x = np.concatenate((x_fill, self.x))
            new_y = np.concatenate((y_fill, self.y))
        elif x_value > x_max:
            x_fill = np.arange(x_max + x_step, x_value + x_step * 0.5, x_step)
            y_fill = np.zeros(x_fill.shape)
            y_fill.fill(y_value)

            new_x = np.concatenate((self.x, x_fill))
            new_y = np.concatenate((self.y, y_fill))
        else:
            return self

        return Pattern(new_x, new_y)

    def to_dict(self) -> dict:
        """
        Returns a dictionary representation of the pattern which can be used to save the pattern to a json file.

        :return: dictionary representation of the pattern
        """
        return {
            'name': self.name,
            'x': self.x.tolist(),
            'y': self.y.tolist(),
            'scaling': self.scaling,
            'offset': self.offset,
            'smoothing': self.smoothing,
            'bkg_pattern': self.bkg_pattern.to_dict() if self.bkg_pattern is
                                                         not None else None
        }

    @staticmethod
    def from_dict(json_dict: dict) -> Pattern:
        """
        Creates a new Pattern from a dictionary representation of a Pattern.

        :param json_dict: dictionary representation of a Pattern
        :return: new Pattern
        """
        pattern = Pattern(np.array(json_dict['x']),
                          np.array(json_dict['y']),
                          json_dict['name'])

        pattern.scaling = json_dict['scaling']
        pattern.offset = json_dict['offset']
        pattern.smoothing = json_dict['smoothing']

        if json_dict['bkg_pattern'] is not None:
            bkg_pattern = Pattern.from_dict(json_dict['bkg_pattern'])
        else:
            bkg_pattern = None
        pattern.bkg_pattern = bkg_pattern

        return pattern

    ###########################################################
    # Operators:

    def __sub__(self, other: Pattern) -> Pattern:
        """
        Subtracts the other pattern from the current one. If the other pattern
        has a different shape, the subtraction will be done on the overlapping
        x-values and the background will be interpolated. If there is no
        overlapping between the two patterns, a BkgNotInRangeError will be
        raised.

        :param other: Pattern to be subtracted
        :return: new Pattern
        """
        orig_x, orig_y = self.data
        other_x, other_y = other.data

        if orig_x.shape != other_x.shape:
            # todo different shape subtraction of spectra seems the fail somehow...
            # the background will be interpolated
            other_fcn = interp1d(other_x, other_y, kind='linear')

            # find overlapping x and y values:
            ind = np.where((orig_x <= np.max(other_x)) &
                           (orig_x >= np.min(other_x)))
            x = orig_x[ind]
            y = orig_y[ind]

            if len(x) == 0:
                # if there is no overlapping between background and pattern, raise an error
                raise BkgNotInRangeError(self.name)
            return Pattern(x, y - other_fcn(x))
        else:
            return Pattern(orig_x, orig_y - other_y)

    def __add__(self, other: Pattern) -> Pattern:
        """
        Adds the other pattern to the current one. If the other pattern
        has a different shape, the addition will be done on the overlapping
        x-values and the y-values of the other pattern will be interpolated. 
        If there is no overlapping between the two patterns, a BkgNotInRangeror 
        will be raised.

        :param other: Pattern to be added
        :return: new Pattern
        """
        orig_x, orig_y = self.data
        other_x, other_y = other.data

        if orig_x.shape != other_x.shape:
            # the background will be interpolated
            other_fcn = interp1d(other_x, other_y, kind='linear')

            # find overlapping x and y values:
            ind = np.where((orig_x <= np.max(other_x)) &
                           (orig_x >= np.min(other_x)))
            x = orig_x[ind]
            y = orig_y[ind]

            if len(x) == 0:
                # if there is no overlapping between background and pattern, raise an error
                raise BkgNotInRangeError(self.name)
            return Pattern(x, y + other_fcn(x))
        else:
            return Pattern(orig_x, orig_y + other_y)

    def __rmul__(self, other: float) -> Pattern:
        """
        Multiplies the pattern with a scalar.

        :param other: scalar to multiply with
        :return: new Pattern
        """
        orig_x, orig_y = self.data
        return Pattern(np.copy(orig_x), np.copy(orig_y) * other)

    def __eq__(self, other: Pattern) -> bool:
        """
        Checks if two patterns are equal. Two patterns are equal if their data
        is equal.

        :param other: Pattern to compare with
        :return: True if equal, False otherwise
        """
        if not isinstance(other, Pattern):
            return False
        if np.array_equal(self.data, other.data):
            return True
        return False


class BkgNotInRangeError(Exception):
    def __init__(self, pattern_name: str):
        self.pattern_name = pattern_name

    def __str__(self):
        return "The background range does not overlap with the Pattern range for " + self.pattern_name



# -*- coding: utf-8 -*/:
from __future__ import annotations
from enum import Enum
from colorsys import hsv_to_rgb
from copy import deepcopy

import numpy as np

from ...core.pattern import Pattern
from ...core.methods import SqMethod, NormalizationMethod, FourierTransformMethod


class TransformConfiguration(object):
    def __init__(self):
        self.use_modification_fcn: bool = False
        self.sq_method: SqMethod = SqMethod.FZ
        self.normalization_method = NormalizationMethod.INTEGRAL
        self.fourier_transform_method = FourierTransformMethod.FFT

        self.q_min: float = 0.0
        self.q_max: float = 10.0

        self.r_min: float = 0.5
        self.r_max: float = 10
        self.r_step: float = 0.01

    def to_dict(self):
        return {
            'use_modification_fcn': self.use_modification_fcn,
            'sq_method': self.sq_method.value,
            'normalization_method': self.normalization_method.value,
            'fourier_transform_method': self.fourier_transform_method.value,
            'q_min': self.q_min,
            'q_max': self.q_max,
            'r_min': self.r_min,
            'r_max': self.r_max,
            'r_step': self.r_step
        }

    @classmethod
    def from_dict(cls, transform_config: dict) -> TransformConfiguration:
        config = cls()
        for key in transform_config.keys():
            if not hasattr(config, key):
                continue
            if key == 'sq_method':
                setattr(config, key, SqMethod(transform_config[key]))
            elif key == 'normalization_method':
                setattr(config, key, NormalizationMethod(transform_config[key]))
            elif key == 'fourier_transform_method':
                setattr(config, key, FourierTransformMethod(transform_config[key]))
            else:
                setattr(config, key, transform_config[key])
        return config


class OptimizeConfiguration(object):
    def __init__(self):
        self.enable = False
        self.r_cutoff = 1.4
        self.iterations = 5
        self.attenuation = 1

    def to_dict(self):
        return vars(self)

    @classmethod
    def from_dict(cls, optimize_config: dict):
        config = cls()
        for key in optimize_config.keys():
            if not hasattr(config, key):
                continue
            setattr(config, key, optimize_config[key])
        return config


class ExtrapolationConfiguration(object):
    def __init__(self):
        self.activate: bool = False
        self.method = 'step'
        self.fit_q_max: float = 2.0
        self.fit_replace: bool = False
        self.s0: float = 0.0
        self.s0_auto: bool = True

    def to_dict(self):
        return vars(self)

    @classmethod
    def from_dict(cls, extrapolation_config: dict):
        config = cls()
        for key in extrapolation_config.keys():
            if not hasattr(config, key):
                continue
            setattr(config, key, extrapolation_config[key])
        return config


class Sample:
    def __init__(self):
        self.sf_source = 'hajdu'
        self.composition = {}
        self.density = 2.2
        self.density_error = None

    def to_dict(self):
        return vars(self)

    @classmethod
    def from_dict(cls, sample_config: dict):
        config = cls()
        for key in sample_config.keys():
            if not hasattr(config, key):
                continue
            setattr(config, key, sample_config[key])
        return config


class SollerConfiguration(object):
    def __init__(self):
        self.enable = False
        self.correction = None
        self.parameters = {'sample_thickness': 1.0,  # in mm
                           'wavelength': 0.31,  # in Angstrom
                           'inner_radius': 62,  # in mm
                           'outer_radius': 210,  # in mm
                           'inner_width': 0.05,  # in mm
                           'outer_width': 0.2,  # in mm
                           'inner_length': 8,  # in mm
                           'outer_length': 6}  # in mm

    def to_dict(self):
        return {'enable': self.enable,
                'correction': self.correction.tolist() if self.correction is not None else None,
                'parameters': self.parameters}

    @classmethod
    def from_dict(cls, soller_config: dict):
        config = cls()
        config.enable = soller_config['enable']
        config.correction = np.array(soller_config['correction']) if soller_config['correction'] is not None else None
        config.parameters = soller_config['parameters']

        return config


class TransferConfiguration(object):
    def __init__(self):
        self.enable = False
        self.function = None
        self.smoothing = 1.0
        self.std_pattern = None
        self.std_bkg_pattern = None
        self.std_bkg_scaling = 1.0
        self.sample_pattern = None
        self.sample_bkg_pattern = None
        self.sample_bkg_scaling = 1.0

    def to_dict(self):
        return {
            'enable': self.enable,
            'smoothing': self.smoothing,
            'std_pattern': self.std_pattern.to_dict() if self.std_pattern is not None else None,
            'std_bkg_pattern': self.std_bkg_pattern.to_dict() if self.std_bkg_pattern is not None else None,
            'std_bkg_scaling': self.std_bkg_scaling,
            'sample_pattern': self.sample_pattern.to_dict() if self.sample_pattern is not None else None,
            'sample_bkg_pattern': self.sample_bkg_pattern.to_dict() if self.sample_bkg_pattern is not None else None,
            'sample_bkg_scaling': self.sample_bkg_scaling,
        }

    @classmethod
    def from_dict(cls, transfer_config: dict):
        config = cls()
        config.enable = transfer_config['enable']
        config.smoothing = transfer_config['smoothing']
        config.std_pattern = Pattern.from_dict(
            transfer_config['std_pattern']) if transfer_config['std_pattern'] is not None else None
        config.std_bkg_pattern = Pattern.from_dict(
            transfer_config['std_bkg_pattern']) if transfer_config['std_bkg_pattern'] is not None else None
        config.std_bkg_scaling = transfer_config['std_bkg_scaling']
        config.sample_pattern = Pattern.from_dict(
            transfer_config['sample_pattern']) if transfer_config['sample_pattern'] is not None else None
        config.sample_bkg_pattern = Pattern.from_dict(
            transfer_config['sample_bkg_pattern']) if transfer_config['sample_bkg_pattern'] is not None else None
        config.sample_bkg_scaling = transfer_config['sample_bkg_scaling']

        return config


class GlassureConfiguration(object):
    num = 0

    def __init__(self):
        super(GlassureConfiguration, self).__init__()
        # initialize all spectra
        self.original_pattern = Pattern()
        self.background_pattern = None

        self.diamond_bkg_pattern = None

        self.sq_pattern = None
        self.fr_pattern = None
        self.gr_pattern = None

        # initialize all parameters
        self.sample = Sample()
        self.transform_config = TransformConfiguration()
        self.optimize_config = OptimizeConfiguration()
        self.extrapolation_config = ExtrapolationConfiguration()
        self.soller_config = SollerConfiguration()
        self.transfer_config = TransferConfiguration()

        self.name = 'Config {}'.format(GlassureConfiguration.num)
        self.color = calculate_color(GlassureConfiguration.num)
        self.show = True
        GlassureConfiguration.num += 1

    def copy(self):
        new_configuration = deepcopy(self)
        new_configuration.name = 'Config {}'.format(GlassureConfiguration.num)
        new_configuration.color = calculate_color(GlassureConfiguration.num)
        GlassureConfiguration.num += 1

        return new_configuration

    def to_dict(self) -> dict:
        config_dict = {
            'original_pattern': self.original_pattern.to_dict(),
            'background_pattern': self.background_pattern.to_dict() if self.background_pattern is not None else None,
            'diamond_bkg_pattern': self.diamond_bkg_pattern.to_dict() if self.diamond_bkg_pattern is not None else None,
            'sq_pattern': self.sq_pattern.to_dict() if self.sq_pattern is not None else None,
            'fr_pattern': self.fr_pattern.to_dict() if self.fr_pattern is not None else None,
            'gr_pattern': self.gr_pattern.to_dict() if self.gr_pattern is not None else None,
            'sample': self.sample.to_dict(),
            'transform_configuration': self.transform_config.to_dict(),
            'optimize_configuration': self.optimize_config.to_dict(),
            'extrapolation_configuration': self.extrapolation_config.to_dict(),
            'soller_configuration': self.soller_config.to_dict(),
            'transfer_configuration': self.transfer_config.to_dict(),
            'name': self.name,
            'color': self.color.tolist(),
            'show': self.show,
        }
        return config_dict

    @classmethod
    def from_dict(cls, config_dict: dict) -> GlassureConfiguration:
        config = cls()

        def get_pattern_or_none(pattern_dict):
            if pattern_dict is None:
                return None
            else:
                return Pattern.from_dict(pattern_dict)

        config.original_pattern = get_pattern_or_none(config_dict['original_pattern'])
        config.background_pattern = get_pattern_or_none(config_dict['background_pattern'])
        config.diamond_bkg_pattern = get_pattern_or_none(config_dict['diamond_bkg_pattern'])
        config.sq_pattern = get_pattern_or_none(config_dict['sq_pattern'])
        config.fr_pattern = get_pattern_or_none(config_dict['fr_pattern'])
        config.gr_pattern = get_pattern_or_none(config_dict['gr_pattern'])

        config.sample = Sample.from_dict(config_dict['sample'])
        config.transform_config = TransformConfiguration.from_dict(config_dict['transform_configuration'])
        config.optimize_config = OptimizeConfiguration.from_dict(config_dict['optimize_configuration'])
        config.extrapolation_config = ExtrapolationConfiguration.from_dict(config_dict['extrapolation_configuration'])
        config.soller_config = SollerConfiguration.from_dict(config_dict['soller_configuration'])
        config.transfer_config = TransferConfiguration.from_dict(config_dict['transfer_configuration'])

        config.name = config_dict['name']
        config.color = np.array(config_dict['color'])
        config.show = config_dict['show'] if 'show' in config_dict.keys() else True

        return config


def calculate_color(ind):
    s = 0.8
    v = 0.8
    h = (0.19 * (ind + 2)) % 1
    return np.array(hsv_to_rgb(h, s, v)) * 255



# -*- coding: utf-8 -*-

import numpy as np
from scipy import interpolate

from .pattern import Pattern
from .utility import convert_density_to_atoms_per_cubic_angstrom, calculate_incoherent_scattering, \
    calculate_f_mean_squared, calculate_f_squared_mean, extrapolate_to_zero_linear

from .calc import calculate_normalization_factor_raw, calculate_sq_raw, calculate_fr, calculate_gr_raw

from .optimization import optimize_sq


class GlassureCalculator(object):
    def __init__(self, original_pattern, background_pattern, composition, density,
                 r=np.linspace(0, 10, 1000)):
        self.original_pattern = original_pattern
        self.background_pattern = background_pattern
        self.sample_pattern = self.original_pattern - self.background_pattern
        self.elemental_abundances = composition
        self.density = density
        self.atomic_density = convert_density_to_atoms_per_cubic_angstrom(composition, density)

        q, _ = self.sample_pattern.data
        self.incoherent_scattering = calculate_incoherent_scattering(composition, q)
        self.f_mean_squared = calculate_f_mean_squared(composition, q)
        self.f_squared_mean = calculate_f_squared_mean(composition, q)

        self.sq_pattern = None
        self.fr_pattern = None
        self.gr_pattern = None

        self.r = r
        self.calculate_transforms(r)

    def calculate_transforms(self, r):
        self.sq_pattern = self.calc_sq()
        self.fr_pattern = self.calc_fr(r)
        self.gr_pattern = self.calc_gr()

    def update_density(self, density):
        self.density = density
        self.atomic_density = convert_density_to_atoms_per_cubic_angstrom(self.elemental_abundances, density)
        self.calculate_transforms()

    def get_normalization_factor(self):
        raise NotImplementedError

    def calc_sq(self):
        raise NotImplementedError

    def calc_fr(self, r):
        raise NotImplementedError

    def calc_gr(self):
        raise NotImplementedError

    def optimize_sq(self, r):
        raise NotImplementedError


class StandardCalculator(GlassureCalculator):
    def __init__(self, original_pattern, background_pattern, composition, density,
                 r=np.linspace(0, 10, 1000), normalization_attenuation_factor=0.001, use_modification_fcn=False,
                 extrapolation_method=None, extrapolation_parameters=None):
        self.attenuation_factor = normalization_attenuation_factor
        self.use_modification_fcn = use_modification_fcn
        self.extrapolation_method = extrapolation_method
        self.extrapolation_parameters = extrapolation_parameters

        super(StandardCalculator, self).__init__(original_pattern, background_pattern,
                                                 composition, density, r)

    def get_normalization_factor(self):
        return calculate_normalization_factor_raw(self.sample_pattern,
                                                  self.atomic_density,
                                                  self.f_squared_mean,
                                                  self.f_mean_squared,
                                                  self.incoherent_scattering,
                                                  self.attenuation_factor)

    def calc_sq(self):
        n = self.get_normalization_factor()
        q, structure_factor = calculate_sq_raw(self.sample_pattern,
                                                self.f_squared_mean,
                                                self.f_mean_squared,
                                                self.incoherent_scattering,
                                                n).data

        if self.extrapolation_method is None:
            return Pattern(q, structure_factor)
        else:
            step = q[1] - q[0]
            q_low = np.arange(step, min(q), step)
            if self.extrapolation_method == 'linear':
                return extrapolate_to_zero_linear(Pattern(q, structure_factor))
            elif self.extrapolation_method == 'spline':
                q_low_cutoff = np.arange(step, self.extrapolation_parameters['cutoff'], step)
                intensity_low_cutoff = np.zeros(q_low_cutoff.shape)

                ind_to_q_max = np.where(q <= self.extrapolation_parameters['q_max'])
                q_spline = np.concatenate((q_low_cutoff, q[ind_to_q_max]))
                int_spline = np.concatenate((intensity_low_cutoff, structure_factor[ind_to_q_max]))

                tck = interpolate.splrep(q_spline, int_spline)

                sq_low = interpolate.splev(q_low, tck)

            return Pattern(np.concatenate((q_low, q)),
                           np.concatenate((sq_low, structure_factor)))

    def calc_fr(self, r=None):
        if r is None:
            r = self.r
        return calculate_fr(self.sq_pattern, r, self.use_modification_fcn)

    def calc_gr(self):
        return calculate_gr_raw(self.fr_pattern, self.atomic_density)

    def optimize_sq(self, r_cutoff, iterations=10, fcn_callback=None, callback_period=2, attenuation_factor=1):
        self.sq_pattern = optimize_sq(self.sq_pattern,
                                       r_cutoff=r_cutoff,
                                       iterations=iterations,
                                       atomic_density=self.atomic_density,
                                       use_modification_fcn=self.use_modification_fcn,
                                       attenuation_factor=attenuation_factor,
                                       fcn_callback=fcn_callback,
                                       callback_period=callback_period)




# -*- coding: utf-8 -*-

import sys
import os


def we_are_frozen():
    # All of the modules are built-in to the interpreter, e.g., by py2exe
    return hasattr(sys, "frozen")


def module_path():
    encoding = sys.getfilesystemencoding()
    if we_are_frozen():
        return os.path.dirname(unicode(sys.executable, encoding))
    return os.path.dirname(unicode(__file__, encoding))


# -*- coding: utf-8 -*-

import sys
import os

from ... import __version__

from qtpy import QtWidgets, QtCore

from .control import CompositionWidget, DataWidget, OptimizationWidget, \
    OptionsWidget, DensityOptimizationWidget, ExtrapolationWidget, \
    DiamondWidget, ConfigurationWidget, SollerWidget, TransferFunctionWidget
from .custom.pattern import PatternWidget
from .custom.box import ExpandableBox


class GlassureWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(GlassureWidget, self).__init__(*args, **kwargs)
        self.horizontal_layout = QtWidgets.QHBoxLayout(self)
        self.horizontal_layout.setContentsMargins(0, 0, 0, 0)
        self.horizontal_layout.setSpacing(0)

        self.pattern_widget = PatternWidget()
        self.left_control_widget = LeftControlWidget()
        self.right_control_widget = RightControlWidget()

        self.left_control_scroll_area = QtWidgets.QScrollArea()
        self.left_control_scroll_area.setWidget(self.left_control_widget)
        self.left_control_scroll_area.setWidgetResizable(True)

        self.left_control_scroll_area.setMaximumWidth(300)
        self.left_control_scroll_area.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)

        self.right_control_scroll_area = QtWidgets.QScrollArea()
        self.right_control_scroll_area.setWidget(self.right_control_widget)
        self.right_control_scroll_area.setWidgetResizable(True)

        self.right_control_scroll_area.setMaximumWidth(300)
        self.right_control_scroll_area.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarAlwaysOff)

        self.horizontal_layout.addWidget(self.left_control_scroll_area)
        self.horizontal_layout.addWidget(self.pattern_widget)
        self.horizontal_layout.addWidget(self.right_control_scroll_area)

        self.horizontal_layout.setStretch(0, 0)
        self.horizontal_layout.setStretch(1, 100)
        self.horizontal_layout.setStretch(2, 0)

        self.setLayout(self.horizontal_layout)

        self.load_stylesheet()

        self.create_widget_shortcuts()
        self.create_function_shortcuts()

        self.setWindowTitle("Glassure v{}".format(__version__))

    def create_widget_shortcuts(self):
        self.load_data_btn = self.left_control_widget.data_widget.file_widget.load_data_btn
        self.load_bkg_btn = self.left_control_widget.data_widget.file_widget.load_background_btn
        self.reset_bkg_btn = self.left_control_widget.data_widget.file_widget.reset_background_btn
        self.data_filename_lbl = self.left_control_widget.data_widget.file_widget.data_filename_lbl
        self.bkg_filename_lbl = self.left_control_widget.data_widget.file_widget.background_filename_lbl

        self.bkg_scaling_sb = self.left_control_widget.data_widget.background_options_gb.scale_sb
        self.bkg_scaling_step_txt = self.left_control_widget.data_widget.background_options_gb.scale_step_txt

        self.smooth_sb = self.left_control_widget.data_widget.smooth_gb.smooth_sb
        self.smooth_step_txt = self.left_control_widget.data_widget.smooth_gb.smooth_step_txt

        self.add_element_btn = self.left_control_widget.composition_widget.add_element_btn
        self.delete_element_btn = self.left_control_widget.composition_widget.delete_element_btn
        self.composition_tw = self.left_control_widget.composition_widget.composition_tw
        self.density_txt = self.left_control_widget.composition_widget.density_txt

        self.q_max_txt = self.left_control_widget.options_widget.q_max_txt
        self.q_min_txt = self.left_control_widget.options_widget.q_min_txt
        self.r_max_txt = self.left_control_widget.options_widget.r_max_txt
        self.r_min_txt = self.left_control_widget.options_widget.r_min_txt
        self.use_modification_cb = self.left_control_widget.options_widget.modification_fcn_cb

        self.activate_extrapolation_cb = self.left_control_widget.extrapolation_widget.activate_cb
        self.extrapolation_q_max_txt = self.left_control_widget.extrapolation_widget.q_max_txt

        self.optimize_activate_cb = self.right_control_widget.optimization_widget.activate_cb
        self.optimize_r_cutoff_txt = self.right_control_widget.optimization_widget.r_cutoff_txt
        self.optimize_iterations_txt = self.right_control_widget.optimization_widget.optimize_iterations_txt
        self.optimize_attenuation_sb = self.right_control_widget.optimization_widget.attenuation_factor_sb

        self.save_sq_btn = self.pattern_widget.mouse_position_widget.save_sq_btn
        self.save_gr_btn = self.pattern_widget.mouse_position_widget.save_gr_btn

        self.configuration_widget = self.right_control_widget.configuration_widget
        self.freeze_configuration_btn = self.right_control_widget.configuration_widget.freeze_btn
        self.remove_configuration_btn = self.right_control_widget.configuration_widget.remove_btn
        self.configuration_tw = self.right_control_widget.configuration_widget.configuration_tw

        self.soller_widget = self.right_control_widget.soller_widget
        self.soller_active_cb = self.right_control_widget.soller_widget.activate_cb

        self.transfer_widget = self.right_control_widget.transfer_widget
        self.transfer_active_cb = self.right_control_widget.transfer_widget.activate_cb

    def create_function_shortcuts(self):
        self.get_sample_config = self.left_control_widget.composition_widget.get_sample_configuration
        self.update_sample_config = self.left_control_widget.composition_widget.update_sample_configuration

        self.get_transform_config = self.left_control_widget.options_widget.get_transform_configuration
        self.update_transform_config = self.left_control_widget.options_widget.update_transform_configuration

        self.get_extrapolation_config = self.left_control_widget.extrapolation_widget.get_configuration
        self.update_extrapolation_config = self.left_control_widget.extrapolation_widget.update_configuration

        self.set_optimization_parameter = self.right_control_widget.optimization_widget.set_parameter
        self.get_optimization_parameter = self.right_control_widget.optimization_widget.get_parameter

        self.set_soller_parameter = self.right_control_widget.soller_widget.set_parameters
        self.get_soller_parameter = self.right_control_widget.soller_widget.get_parameters

    def show(self):
        QtWidgets.QWidget.show(self)
        if sys.platform == "darwin":
            self.setWindowState(
                self.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
            self.activateWindow()
            self.raise_()

    def load_stylesheet(self):
        stylesheet_file = open(os.path.join(
            module_path(), "DioptasStyle.qss"), 'r')
        stylesheet_str = stylesheet_file.read()
        self.setStyleSheet(stylesheet_str)


class LeftControlWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(LeftControlWidget, self).__init__(*args, **kwargs)
        self.vertical_layout = QtWidgets.QVBoxLayout()
        self.vertical_layout.setSpacing(8)
        self.vertical_layout.setContentsMargins(5, 5, 5, 5)

        self.data_widget = DataWidget()
        self.composition_widget = CompositionWidget()
        self.options_widget = OptionsWidget()
        self.extrapolation_widget = ExtrapolationWidget()

        self.vertical_layout.addWidget(ExpandableBox(self.data_widget, "Data"))
        self.vertical_layout.addWidget(ExpandableBox(self.composition_widget, "Composition"))
        self.vertical_layout.addWidget(ExpandableBox(self.options_widget, "Options"))
        self.vertical_layout.addWidget(ExpandableBox(self.extrapolation_widget, "Extrapolation"))

        self.vertical_layout.addSpacerItem(
            QtWidgets.QSpacerItem(20, 50,
                                  QtWidgets.QSizePolicy.Fixed,
                                  QtWidgets.QSizePolicy.Expanding))

        self.setLayout(self.vertical_layout)


class RightControlWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(RightControlWidget, self).__init__(*args, **kwargs)
        self.vertical_layout = QtWidgets.QVBoxLayout()
        self.vertical_layout.setSpacing(8)
        self.vertical_layout.setContentsMargins(5, 5, 5, 5)

        self.configuration_widget = ConfigurationWidget()
        self.optimization_widget = OptimizationWidget()
        self.density_optimization_widget = DensityOptimizationWidget()
        self.diamond_widget = DiamondWidget()
        self.soller_widget = SollerWidget()
        self.transfer_widget = TransferFunctionWidget()

        self.vertical_layout.addWidget(ExpandableBox(
            self.configuration_widget, "Configurations"))
        self.vertical_layout.addWidget(ExpandableBox(
            self.optimization_widget, "Optimization"))
        self.vertical_layout.addWidget(ExpandableBox(
            self.density_optimization_widget, "Density Optimization", True))
        self.vertical_layout.addWidget(ExpandableBox(
            self.diamond_widget, "Diamond Correction", True))
        self.vertical_layout.addWidget(ExpandableBox(
            self.soller_widget, "Soller Slit Correction", True))
        self.vertical_layout.addWidget(ExpandableBox(
            self.transfer_widget, "Transfer Function Correction", True))

        self.vertical_layout.addSpacerItem(
            QtWidgets.QSpacerItem(20, 50, QtWidgets.QSizePolicy.Fixed,
                                  QtWidgets.QSizePolicy.MinimumExpanding))

        self.setLayout(self.vertical_layout)


def we_are_frozen():
    # All of the modules are built-in to the interpreter, e.g., by py2exe
    return hasattr(sys, "frozen")


def module_path():
    return os.path.dirname(__file__)



from enum import Enum


class SqMethod(Enum):
    """
    Enum class for the different methods to calculate the structure factor.
    """
    FZ = 'FZ'
    AL = 'AL'


class NormalizationMethod(Enum):
    """
    Enum class for the different methods to perform an intensity normalization.
    """
    INTEGRAL = 'integral'
    FIT = 'fit'


class FourierTransformMethod(Enum):
    """
    Enum class for the different methods to perform a Fourier transform.
    """
    FFT = 'fft'
    INTEGRAL = 'integral'



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets

from ..custom import FloatLineEdit


class DataWidget(QtWidgets.QWidget):
    def __init__(self):
        super(DataWidget, self).__init__()

        self.file_widget = FileWidget()
        self.background_options_gb = BackgroundOptionsGroupBox()
        self.smooth_gb = SmoothGroupBox()

        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.file_widget)
        self._layout.addWidget(self.background_options_gb)
        self._layout.addWidget(self.smooth_gb)
        self._layout.addSpacing(5)

        self.setLayout(self._layout)


class FileWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(FileWidget, self).__init__(*args, **kwargs)

        self.vertical_layout = QtWidgets.QVBoxLayout()
        self.vertical_layout.setContentsMargins(0, 0, 0, 0)

        self.load_data_btn = QtWidgets.QPushButton("Load Data")
        self.data_filename_lbl = QtWidgets.QLabel("None")
        self.data_filename_lbl.setAlignment(QtCore.Qt.AlignRight)

        self.load_background_btn = QtWidgets.QPushButton("Load Bkg")
        # use icon for reset button
        pixmapi = getattr(QtWidgets.QStyle, "SP_DialogCancelButton")
        icon = self.style().standardIcon(pixmapi)
        self.reset_background_btn = QtWidgets.QPushButton(icon, "")

        self.background_filename_lbl = QtWidgets.QLabel("None")
        self.background_filename_lbl.setAlignment(QtCore.Qt.AlignRight)

        self.vertical_layout.addWidget(self.load_data_btn)
        self.vertical_layout.addWidget(self.data_filename_lbl)

        self.background_btn_layout = QtWidgets.QHBoxLayout()
        self.background_btn_layout.addWidget(self.load_background_btn)
        self.background_btn_layout.addWidget(self.reset_background_btn)

        self.vertical_layout.addLayout(self.background_btn_layout)             
        self.vertical_layout.addWidget(self.background_filename_lbl)

        self.setLayout(self.vertical_layout)
        self.style_widgets()

    def style_widgets(self):
        self.reset_background_btn.setMaximumWidth(20)
        self.reset_background_btn.setMaximumHeight(20)
        self.reset_background_btn.setToolTip("Reset Background")


class BackgroundOptionsGroupBox(QtWidgets.QGroupBox):
    def __init__(self, *args):
        super(BackgroundOptionsGroupBox, self).__init__(*args)

        self.create_widgets()
        self.style_widgets()
        self.create_layout()
        self.create_signals()

    def create_widgets(self):
        self.scale_lbl = QtWidgets.QLabel("Scale:")
        self.scale_sb = QtWidgets.QDoubleSpinBox()
        self.scale_step_txt = FloatLineEdit("0.01")

    def style_widgets(self):
        self.scale_lbl.setAlignment(QtCore.Qt.AlignVCenter |
                                    QtCore.Qt.AlignRight)

        self.scale_sb.setValue(1.0)
        self.scale_sb.setSingleStep(0.01)
        self.scale_sb.setDecimals(4)
        self.scale_sb.setMinimumWidth(80)

        self.scale_step_txt.setMaximumWidth(60)
        self.scale_sb.setAlignment(QtCore.Qt.AlignRight)
        self.scale_step_txt.setAlignment(QtCore.Qt.AlignRight)

    def create_layout(self):
        self.grid_layout = QtWidgets.QGridLayout()
        self.grid_layout.setContentsMargins(3, 5, 5, 3)
        self.grid_layout.setSpacing(5)

        self.grid_layout.addWidget(self.scale_lbl, 0, 0)
        self.grid_layout.addWidget(self.scale_sb, 0, 1)
        self.grid_layout.addWidget(self.scale_step_txt, 0, 2)

        self.setLayout(self.grid_layout)

    def create_signals(self):
        self.scale_step_txt.editingFinished.connect(self.scale_step_changed)

    def scale_step_changed(self):
        self.scale_sb.setSingleStep(float(str(self.scale_step_txt.text())))


class SmoothGroupBox(QtWidgets.QGroupBox):
    def __init__(self, *args):
        super(SmoothGroupBox, self).__init__(*args)
        self.grid_layout = QtWidgets.QGridLayout()
        self.grid_layout.setContentsMargins(3, 5, 5, 3)
        self.grid_layout.setSpacing(5)

        self.smooth_lbl = QtWidgets.QLabel("Smooth:")
        self.smooth_lbl.setAlignment(QtCore.Qt.AlignVCenter |
                                     QtCore.Qt.AlignRight)

        self.smooth_sb = QtWidgets.QDoubleSpinBox()
        self.smooth_sb.setAlignment(QtCore.Qt.AlignRight)
        self.smooth_sb.setSingleStep(1)
        self.smooth_sb.setMinimumWidth(80)

        self.smooth_step_txt = FloatLineEdit("1")
        self.smooth_step_txt.setMaximumWidth(60)

        self.smooth_step_txt.editingFinished.connect(self.step_changed)

        self.grid_layout.addWidget(self.smooth_lbl, 0, 0)
        self.grid_layout.addWidget(self.smooth_sb, 0, 1)
        self.grid_layout.addWidget(self.smooth_step_txt, 0, 2)

        self.setLayout(self.grid_layout)

    def step_changed(self):
        self.smooth_sb.setSingleStep(float(str(self.smooth_step_txt.text())))



# -*- coding: utf-8 -*-

from functools import partial

from qtpy import QtCore, QtWidgets, QtGui
from ..custom import FlatButton, ListTableWidget


class ConfigurationWidget(QtWidgets.QWidget):
    configuration_color_btn_clicked = QtCore.Signal(int, QtWidgets.QWidget)
    configuration_show_cb_state_changed = QtCore.Signal(int, bool)
    configuration_name_changed = QtCore.Signal(int, str)

    def __init__(self, *args):
        super(ConfigurationWidget, self).__init__(*args)

        self._create_widgets()
        self._create_layout()
        self._style_widgets()

        self.configuration_show_cbs = []
        self.configuration_color_btns = []

    def _create_widgets(self):
        self.freeze_btn = QtWidgets.QPushButton("Freeze")
        self.remove_btn = QtWidgets.QPushButton("Remove")
        self.configuration_tw = ListTableWidget(columns=3)
        self.configuration_tw.setObjectName('configuration_tw')
        self.configuration_tw.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        # First column will not trigger selection of a row
        self.configuration_tw.setItemDelegateForColumn(0, QtWidgets.QStyledItemDelegate())

        self.configuration_tw.cellChanged.connect(self.configuration_label_editingFinished)

        pixmap = getattr(QtWidgets.QStyle, "SP_DialogSaveButton")
        icon = self.style().standardIcon(pixmap)
        self.save_btn = QtWidgets.QPushButton(icon, "")
        pixmap = getattr(QtWidgets.QStyle, "SP_DialogOpenButton")
        icon = self.style().standardIcon(pixmap)
        self.load_btn = QtWidgets.QPushButton(icon, "")

    def _create_layout(self):
        self._button_layout = QtWidgets.QHBoxLayout()
        self._button_layout.addWidget(self.freeze_btn)
        self._button_layout.addWidget(self.remove_btn)

        self._main_layout = QtWidgets.QVBoxLayout()
        self._main_layout.addLayout(self._button_layout)
        self._main_layout.addWidget(self.configuration_tw)

        self._save_load_layout = QtWidgets.QHBoxLayout()
        self._save_load_layout.addStretch()
        self._save_load_layout.addWidget(self.load_btn)
        self._save_load_layout.addWidget(self.save_btn)
        self._main_layout.addLayout(self._save_load_layout)

        self.setLayout(self._main_layout)

    def _style_widgets(self):
        self.setStyleSheet("""
            #configuration_tw QPushButton {
                margin-top: 4;
                max-width: 7;
                max-height: 16;
            }
        """)
        self.save_btn.setMaximumWidth(23)
        self.save_btn.setMinimumWidth(23)
        self.save_btn.setMaximumHeight(23)
        self.save_btn.setMinimumHeight(23)
        self.save_btn.setToolTip("Save all configurations into a json file")
        self.load_btn.setMaximumWidth(23)
        self.load_btn.setMaximumHeight(23)
        self.load_btn.setMinimumWidth(23)
        self.load_btn.setMinimumHeight(23)
        self.load_btn.setToolTip(
            "Load multiple configurations from a json file")
        self._save_load_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.setContentsMargins(0, 0, 0, 0)

    def add_configuration(self, name, color, show=True):
        self.configuration_tw.blockSignals(True)
        current_rows = self.configuration_tw.rowCount()
        self.configuration_tw.setRowCount(current_rows + 1)

        show_cb = QtWidgets.QCheckBox()
        show_cb.setChecked(show)
        show_cb.stateChanged.connect(partial(self.configuration_show_cb_changed, show_cb))
        show_cb.setStyleSheet("background-color: transparent")
        self.configuration_tw.setCellWidget(current_rows, 0, show_cb)
        self.configuration_show_cbs.append(show_cb)

        color_button = FlatButton()
        color_button.setStyleSheet("background-color: " + color)
        color_button.clicked.connect(partial(self.configuration_color_btn_click, color_button))
        self.configuration_tw.setCellWidget(current_rows, 1, color_button)
        self.configuration_color_btns.append(color_button)

        name_item = QtWidgets.QTableWidgetItem(name)
        name_item.setFlags(name_item.flags() & ~QtCore.Qt.ItemIsEditable)
        self.configuration_tw.setItem(current_rows, 2, QtWidgets.QTableWidgetItem(name))

        self.configuration_tw.setColumnWidth(0, 20)
        self.configuration_tw.setColumnWidth(1, 25)
        self.configuration_tw.setRowHeight(current_rows, 25)
        self.select_configuration(current_rows)
        self.configuration_tw.blockSignals(False)

    def select_configuration(self, ind):
        if self.configuration_tw.rowCount() > 0:
            self.configuration_tw.selectRow(ind)

    def get_selected_configuration_row(self):
        selected = self.configuration_tw.selectionModel().selectedRows()
        try:
            row = selected[0].row()
        except IndexError:
            row = -1
        return row

    def remove_configuration(self, ind):
        if self.configuration_tw.rowCount() == 0:
            return

        self.configuration_tw.blockSignals(True)
        self.configuration_tw.removeRow(ind)
        self.configuration_tw.blockSignals(False)

        self.configuration_show_cbs.remove(self.configuration_show_cbs[ind])
        self.configuration_color_btns.remove(
            self.configuration_color_btns[ind])

        if self.configuration_tw.rowCount() > ind:
            self.select_configuration(ind)
        else:
            self.select_configuration(self.configuration_tw.rowCount() - 1)

    def clear_configuration_tw(self):
        self.configuration_tw.clear()
        self.configuration_tw.setRowCount(0)

        self.configuration_show_cbs = []
        self.configuration_color_btns = []

    def configuration_color_btn_click(self, button):
        self.configuration_color_btn_clicked.emit(
            self.configuration_color_btns.index(button), button)

    def configuration_show_cb_changed(self, checkbox, _):
        self.configuration_show_cb_state_changed.emit(self.configuration_show_cbs.index(checkbox),
                                                      checkbox.isChecked())

    def configuration_show_cb_set_checked(self, ind, state):
        checkbox = self.configuration_show_cbs[ind]
        checkbox.setChecked(state)

    def configuration_show_cb_is_checked(self, ind):
        checkbox = self.configuration_show_cbs[ind]
        return checkbox.isChecked()

    def configuration_label_editingFinished(self, row, col):
        label_item = self.configuration_tw.item(row, col)
        self.configuration_name_changed.emit(int(row), str(label_item.text()))



# -*- coding: utf-8 -*-
from __future__ import annotations
import pyqtgraph as pg
import numpy as np
from qtpy import QtCore, QtWidgets

from glassure.core.pattern import Pattern

Signal = QtCore.Signal


class PatternWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(PatternWidget, self).__init__(*args, **kwargs)
        self._layout = QtWidgets.QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(8)
        self.create_plots()
        self.style_plots()
        self.create_items()

        self.mouse_position_widget = MousePositionWidget()
        self._layout.addWidget(self.mouse_position_widget)

        self.create_signals()

        self.setLayout(self._layout)

    def create_plots(self):
        self.pg_layout_widget = pg.GraphicsLayoutWidget()
        self.pg_layout = pg.GraphicsLayout()
        self.pg_layout.setContentsMargins(0, 0, 0, 0)
        self.pg_layout_widget.setContentsMargins(0, 0, 0, 0)

        self.pattern_plot = ModifiedPlotItem()
        self.sq_plot = ModifiedPlotItem()
        self.gr_plot = ModifiedPlotItem()

        self.pg_layout.addItem(self.pattern_plot, 0, 0)
        self.pg_layout.addItem(self.sq_plot, 1, 0)
        self.pg_layout.addItem(self.gr_plot, 2, 0)

        self.pg_layout_widget.addItem(self.pg_layout)

        self._layout.addWidget(self.pg_layout_widget)

    def style_plots(self):
        self.pattern_plot.setLabel('bottom', text='Q (1/A)')
        self.pattern_plot.setLabel('left', text='Int (a.u.)')

        self.sq_plot.setLabel('bottom', text='Q (1/A)')
        self.sq_plot.setLabel('left', text='S(Q)')

        self.gr_plot.setLabel('bottom', text='r (A)')
        self.gr_plot.setLabel('left', text='g(r)')

    def create_items(self):
        self.pattern_item = pg.PlotDataItem(pen=pg.mkPen('w', width=1.5))
        self.bkg_item = pg.PlotDataItem(pen=pg.mkPen(
            'r', width=1.5, style=QtCore.Qt.DashLine))
        self.pattern_plot.addItem(self.pattern_item)
        self.pattern_plot.addItem(self.bkg_item)

        self.sq_items = []
        self.sq_show = []
        self.add_sq_item()

        self.gr_items = []
        self.gr_show = []
        self.add_gr_item()

    def add_sq_item(self, color='w', show=True):
        self.sq_items.append(pg.PlotDataItem([], [], pen=pg.mkPen(color=color, width=1.5)))
        self.sq_show.append(show)
        if show:
            self.sq_plot.addItem(self.sq_items[-1])

    def add_gr_item(self, color='w', show=True):
        self.gr_items.append(pg.PlotDataItem([], [], pen=pg.mkPen(color=color, width=1.5)))
        self.gr_show.append(show)
        if show:
            self.gr_plot.addItem(self.gr_items[-1])

    def remove_sq_item(self, ind=-1):
        self.sq_plot.removeItem(self.sq_items[ind])
        del self.sq_items[ind]
        del self.sq_show[ind]

    def remove_gr_item(self, ind=-1):
        self.gr_plot.removeItem(self.gr_items[ind])
        del self.gr_items[ind]
        del self.gr_show[ind]

    def create_signals(self):
        self.pattern_plot.connect_mouse_move_event()
        self.sq_plot.connect_mouse_move_event()
        self.gr_plot.connect_mouse_move_event()
        self.pattern_plot.mouse_moved.connect(self.mouse_moved)
        self.sq_plot.mouse_moved.connect(self.mouse_moved)
        self.gr_plot.mouse_moved.connect(self.mouse_moved)

    def mouse_moved(self, x, y):
        self.mouse_position_widget.x_value_lbl.setText("{:9.3f}".format(x))
        self.mouse_position_widget.y_value_lbl.setText("{:9.3f}".format(y))

    def plot_pattern(self, pattern):
        x, y = pattern.data
        self.pattern_item.setData(x=x, y=y)

    def plot_bkg(self, pattern: Pattern | None):
        if pattern is None:
            self.bkg_item.setData(x=[], y=[])
        else:
            x, y = pattern.data
            self.bkg_item.setData(x=x, y=y)

    def set_sq_pattern(self, pattern, ind):
        x, y = pattern.data
        self.sq_items[ind].setData(x=x, y=y)

    def set_gr_pattern(self, pattern, ind):
        x, y = pattern.data
        self.gr_items[ind].setData(x=x, y=y)

    def set_color(self, color, ind):
        self.sq_items[ind].setPen(pg.mkPen(color=color, width=1.5))
        self.gr_items[ind].setPen(pg.mkPen(color=color, width=1.5))
        self.sq_items[ind].setZValue(0)
        self.gr_items[ind].setZValue(0)

    def activate_ind(self, ind):
        self.sq_items[ind].setPen(pg.mkPen(color='w', width=2))
        self.gr_items[ind].setPen(pg.mkPen(color='w', width=2))
        self.sq_items[ind].setZValue(100)
        self.gr_items[ind].setZValue(100)

    def hide_sq(self, ind):
        self.sq_plot.removeItem(self.sq_items[ind])
        self.sq_show[ind] = False

    def hide_gr(self, ind):
        self.gr_plot.removeItem(self.gr_items[ind])
        self.gr_show[ind] = False

    def show_sq(self, ind):
        self.sq_plot.addItem(self.sq_items[ind])
        self.sq_show[ind] = True

    def show_gr(self, ind):
        self.gr_plot.addItem(self.gr_items[ind])
        self.gr_show[ind] = True

    def set_configuration_color(self, color, ind):
        self.sq_items[ind].setPen(pg.mkPen(color=color, width=1.5))
        self.gr_items[ind].setPen(pg.mkPen(color=color, width=1.5))


class ModifiedPlotItem(pg.PlotItem):
    mouse_moved = Signal(float, float)
    mouse_left_clicked = Signal(float, float)
    range_changed = Signal(list)

    def __init__(self, *args, **kwargs):
        super(ModifiedPlotItem, self).__init__(*args, **kwargs)

        self.modify_mouse_behavior()

    def modify_mouse_behavior(self):
        self.vb.mouseClickEvent = self.mouse_click_event
        self.vb.mouseDragEvent = self.mouse_drag_event
        self.vb.mouseDoubleClickEvent = self.mouse_double_click_event
        self.vb.wheelEvent = self.wheel_event
        self.range_changed_timer = QtCore.QTimer()
        self.range_changed_timer.timeout.connect(self.emit_sig_range_changed)
        self.range_changed_timer.setInterval(30)
        self.last_view_range = np.array(self.vb.viewRange())

    def connect_mouse_move_event(self):
        self.scene().sigMouseMoved.connect(self.mouse_move_event)

    def mouse_move_event(self, pos):
        if self.sceneBoundingRect().contains(pos):
            pos = self.vb.mapSceneToView(pos)
            self.mouse_moved.emit(pos.x(), pos.y())

    def mouse_click_event(self, ev):
        if ev.button() == QtCore.Qt.RightButton or \
                (ev.button() == QtCore.Qt.LeftButton and
                 ev.modifiers() & QtCore.Qt.ControlModifier):
            self.vb.scaleBy((2, 2))
            self.vb.sigRangeChangedManually.emit(self.vb.state['mouseEnabled'])
        elif ev.button() == QtCore.Qt.LeftButton:
            if self.sceneBoundingRect().contains(ev.pos()):
                pos = self.vb.mapToView(ev.pos())
                x = pos.x()
                y = pos.y()
                self.mouse_left_clicked.emit(x, y)

    def mouse_double_click_event(self, ev):
        if (ev.button() == QtCore.Qt.RightButton) or \
                (ev.button() == QtCore.Qt.LeftButton and
                 ev.modifiers() & QtCore.Qt.ControlModifier):
            self.vb.autoRange()
            self.vb.enableAutoRange()
            self._auto_range = True
            self.vb.sigRangeChangedManually.emit(self.vb.state['mouseEnabled'])

    def mouse_drag_event(self, ev, axis=None):
        # most of this code is copied behavior mouse drag from the original code
        ev.accept()
        pos = ev.pos()
        last_pos = ev.lastPos()
        dif = pos - last_pos
        dif *= -1

        if ev.button() == QtCore.Qt.RightButton or \
                (ev.button() == QtCore.Qt.LeftButton and
                 ev.modifiers() & QtCore.Qt.ControlModifier):
            # determine the amount of translation
            tr = dif
            tr = self.vb.mapToView(tr) - self.vb.mapToView(pg.Point(0, 0))
            x = tr.x()
            y = tr.y()
            self.vb.translateBy(x=x, y=y)
            if ev.start:
                self.range_changed_timer.start()
            if ev.isFinish():
                self.range_changed_timer.stop()
                self.emit_sig_range_changed()
        else:
            if ev.isFinish():
                # This is the final move in the drag; change the view scale now
                self._auto_range = False
                self.vb.enableAutoRange(enable=False)
                self.vb.rbScaleBox.hide()
                ax = QtCore.QRectF(
                    pg.Point(ev.buttonDownPos(ev.button())), pg.Point(pos))
                ax = self.vb.childGroup.mapRectFromParent(ax)
                self.vb.showAxRect(ax)
                self.vb.axHistoryPointer += 1
                self.vb.axHistory = self.vb.axHistory[
                                    :self.vb.axHistoryPointer] + [ax]
                self.vb.sigRangeChangedManually.emit(
                    self.vb.state['mouseEnabled'])
            else:
                # update shape of scale box
                self.vb.updateScaleBox(ev.buttonDownPos(), ev.pos())

    def emit_sig_range_changed(self):
        new_view_range = np.array(self.vb.viewRange())
        if not np.array_equal(self.last_view_range, new_view_range):
            self.vb.sigRangeChangedManually.emit(self.vb.state['mouseEnabled'])
            self.last_view_range = new_view_range

    def wheel_event(self, ev, axis=None, *args):
        pg.ViewBox.wheelEvent(self.vb, ev, axis)
        self.vb.sigRangeChangedManually.emit(self.vb.state['mouseEnabled'])


class MousePositionWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(MousePositionWidget, self).__init__()

        self.horizontal_layout = QtWidgets.QHBoxLayout()
        self.horizontal_layout.setContentsMargins(0, 0, 0, 5)
        self.horizontal_layout.setSpacing(5)
        self.x_unit_lbl = QtWidgets.QLabel('x:')
        self.x_value_lbl = QtWidgets.QLabel('0.00')

        self.y_unit_lbl = QtWidgets.QLabel('y:')
        self.y_value_lbl = QtWidgets.QLabel('0.00')

        self.x_unit_lbl.setAlignment(
            QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.x_value_lbl.setAlignment(
            QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.y_unit_lbl.setAlignment(
            QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)
        self.y_value_lbl.setAlignment(
            QtCore.Qt.AlignVCenter | QtCore.Qt.AlignRight)

        self.x_unit_lbl.setFixedWidth(30)
        self.y_unit_lbl.setFixedWidth(30)

        self.x_value_lbl.setFixedWidth(50)
        self.y_value_lbl.setFixedWidth(50)

        self.save_sq_btn = QtWidgets.QPushButton("Save S(Q)")
        self.save_sq_btn.setFlat(True)
        self.save_gr_btn = QtWidgets.QPushButton("Save g(r)")
        self.save_gr_btn.setFlat(True)

        self.horizontal_layout.addWidget(self.x_unit_lbl)
        self.horizontal_layout.addWidget(self.x_value_lbl)
        self.horizontal_layout.addWidget(self.y_unit_lbl)
        self.horizontal_layout.addWidget(self.y_value_lbl)
        self.horizontal_layout.addSpacerItem(
            QtWidgets.QSpacerItem(10, 10,
                                  QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Fixed))

        self.horizontal_layout.addWidget(self.save_sq_btn)
        self.horizontal_layout.addWidget(self.save_gr_btn)
        self.setLayout(self.horizontal_layout)



# -*- coding: utf-8 -*-

from qtpy import QtWidgets


def HorizontalLine():
    frame = QtWidgets.QFrame()
    frame.setFrameShape(QtWidgets.QFrame.HLine)
    frame.setStyleSheet("border: 2px solid #CCC;")
    frame.setFrameShadow(QtWidgets.QFrame.Sunken)
    return frame



# -*- coding: utf-8 -*-

from qtpy import QtCore, QtGui, QtWidgets
from .lines import HorizontalLine

Signal = QtCore.Signal


def VerticalSpacerItem():
    return QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum,
                                 QtWidgets.QSizePolicy.Expanding)


def HorizontalSpacerItem():
    return QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Expanding,
                                 QtWidgets.QSizePolicy.Ex)


class CommaDoubleValidator(QtGui.QDoubleValidator):
    """
    This class is used to validate the number input of a NumberTextfield
    widget. It allows the user to input a number with a comma or a dot as
    decimal separator.
    """

    def validate(self, text, pos):
        text = text.replace(",", ".")
        return super(CommaDoubleValidator, self).validate(text, pos)


class FloatLineEdit(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super(FloatLineEdit, self).__init__(*args, **kwargs)
        self.setValidator(CommaDoubleValidator())
        self.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

    def text(self):
        return super(FloatLineEdit, self).text().replace(",", ".")

    def value(self):
        return float(self.text())


class IntegerLineEdit(QtWidgets.QLineEdit):
    def __init__(self, *args, **kwargs):
        super(IntegerLineEdit, self).__init__(*args, **kwargs)
        self.setValidator(QtGui.QIntValidator())
        self.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)

    def value(self):
        return int(self.text())


class LabelAlignRight(QtWidgets.QLabel):
    def __init__(self, *args, **kwargs):
        super(LabelAlignRight, self).__init__(*args, **kwargs)
        self.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)


class FlatButton(QtWidgets.QPushButton):
    def __init__(self, *args):
        super(FlatButton, self).__init__(*args)
        self.setFlat(True)

    def text(self):
        return super(FloatLineEdit, self).text().replace(",", ".")

    def value(self):
        return float(self.text())


class CheckableFlatButton(QtWidgets.QPushButton):
    def __init__(self, *args):
        super(CheckableFlatButton, self).__init__(*args)
        self.setFlat(True)
        self.setCheckable(True)


class ListTableWidget(QtWidgets.QTableWidget):
    def __init__(self, columns=3):
        super(ListTableWidget, self).__init__()

        self.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.setColumnCount(columns)
        self.horizontalHeader().setVisible(False)
        self.verticalHeader().setVisible(False)
        self.horizontalHeader().setStretchLastSection(True)
        self.setShowGrid(False)


class ValueLabelTxtPair(QtWidgets.QWidget):
    editingFinished = Signal()

    def __init__(self, label_str, value_init, unit_str, layout, layout_row=0, layout_col=0, parent=None):
        super(ValueLabelTxtPair, self).__init__(parent)

        self.desc_lbl = LabelAlignRight(label_str)
        self.value_txt = FloatLineEdit(str(value_init))
        self.unit_lbl = LabelAlignRight(unit_str)

        self.layout = layout

        self.layout.addWidget(self.desc_lbl, layout_row, layout_col)
        self.layout.addWidget(self.value_txt, layout_row, layout_col + 1)
        self.layout.addWidget(self.unit_lbl, layout_row, layout_col + 2)

        self.value_txt.editingFinished.connect(self.editingFinished.emit)

    def get_value(self):
        return float(str(self.value_txt.text()))

    def set_value(self, value):
        self.value_txt.setText(str(value))

    def setText(self, new_str):
        self.value_txt.setText(new_str)


class DragSlider(QtWidgets.QSlider):
    dragChanged = Signal(int)

    def __init__(self, parent=None):
        super(DragSlider, self).__init__(parent)

        self.setRange(-100, 100)
        self.setSingleStep(1)
        self.setPageStep(0)

        self._drag_value = 0

        self.setOrientation(QtCore.Qt.Horizontal)

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.send_value)

        self.sliderPressed.connect(self.start_drag)
        self.sliderReleased.connect(self.stop_drag)
        super(DragSlider, self).valueChanged.connect(self._update_value)

    def _update_value(self, value):
        self._drag_value = value

    def reset_value(self):
        self._drag_value = 0

    def start_drag(self):
        if not self.timer.isActive():
            self.timer.start(100)

    def send_value(self):
        self.dragChanged.emit(self._drag_value)

    def stop_drag(self):
        self.timer.stop()
        self.blockSignals(True)
        self.setValue(0)
        self.blockSignals(False)
